#!/usr/bin/env python3
# -*- coding: utf8 -*-

# Copyright 2024 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from omflut import FlutterGroup
from omflut import FakeStructBuilder
from omflut import RbfXferBuilder
from sdpm.api.om import SdpmBuilder
from pypk.api_om import PypkBuilder
import openmdao.api as om
import numpy as np

def get_path(fname):
    import os.path
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data', fname)

def get_exp():
    mach = [0.499, 0.678, 0.901, 0.960] # experimental Mach number
    rho = [0.4278, 0.2082, 0.0995, 0.0634] # experimental density
    mu = [33.465, 68.753, 143.920, 225.820] # experimental mass ratio m/(rho*V)
    n = len(mach)
    return n, mach, rho, mu

def get_cfg(irun):
    import os.path
    # Shared parameters
    c_ref = 0.559 # root chord
    k_ref = [0.01, 0.1, 0.2, 0.3] # reference reduced frequencies
    n_mod = 4 # number of mode
    # Experimental Mach number, density and mass ratio
    _, mach, rho, mu = get_exp()
    # Structural config
    s_cfg = {
        'File': get_path('agard445_modes.csv'), # input file containing the modes
        'M': np.diag([1.0, 1.0, 1.0, 1.0]), # mass matrix
        'K': np.diag([3695.2, 63392.2, 99892.1, 368026.0]) # stifness matrix
    }
    # SDPM config
    wnc = 20
    a_cfg = {
        # Verbosity
        'Verb_lvl': 0,
        # Model (geometry or mesh)
        'File': get_path('agard445.geo'), # input Gmsh file
        'Pars': {'nC': wnc, 'nS': 10, 'pC': 1.1, 'pS': 1.0}, # parameters for Gmsh file (optional)
        # Markers
        'Wing': 'wing', # name of the marker defining the lifting body (wing)
        'Te': 'te', # name of the marker defining the trailing edge of the lifting body
        # Freestream
        'Mach': mach[irun], # Mach number (optional)
        'AoA': 0.0, # angle of attack (optional)
        'AoS': 0.0, # angle of side slip (optional)
        # Unsteady motion
        'Num_wake_div': wnc, # number of cells per chord length in the wake
        'Frequencies': k_ref, # list of reduced frequencies
        'Num_modes': n_mod, # number of modes
        # Geometry
        'Symmetry': True, # whether only half (symmetric) model is provided
        's_ref': 0.353, # reference surface area
        'c_ref': c_ref, # reference chord length
        'x_ref': 0.0, # reference point for moment computation (x)
        'y_ref': 0.0, # reference point for moment computation (y)
        'z_ref': 0.0 # reference point for moment computation (z)
    }
    # RBF
    x_cfg = {
    'n_dim': 2, # number of coordinates to use for interpolation
    'n_modes': n_mod, # number of modes
    'smooth': 1 # smoothing
    }
    # NIPK
    f_cfg = {
        'name': 'agard', # case name
        'k_ref': np.array(k_ref), # reference reduced frequencies
        'l_ref': 0.5 * c_ref, # reference length (half root chord)
        'mach': mach[irun], # freesteam Mach number
        'rho_inf': rho[irun], # freesteam density
        'u_idx': np.linspace(0.2, 0.5, 31, endpoint=True), # velocity index range
        'mu': mu[irun], # mass ratio
        'f_ref': 40.3511, # frequency of first torsional mode
        'n_modes': n_mod, # number of modes
        'rho_ks': 1e6, # KS aggregation parameter for KS
        'method': 'nipk', # method type
        'fluid': 'unmatched', # fluid type
        'vrb': 1 # verbosity level
    }
    return s_cfg, a_cfg, x_cfg, f_cfg

class Agard(om.Group):
    """Aerostructural flutter model of the AGARD 445.6 wing
    """
    def initialize(self):
        self.options.declare('n', default=1, desc='number of runs')

    def setup(self):
        for i in range(self.options['n']):
            # Builders
            s_cfg, a_cfg, x_cfg, f_cfg = get_cfg(i)
            agrd = FakeStructBuilder(s_cfg)
            sdpm = SdpmBuilder(a_cfg)
            rbfi = RbfXferBuilder(agrd, sdpm, x_cfg)
            pypk = PypkBuilder(f_cfg)
            # Flutter component
            self.add_subsystem(f'flutter_{i}', FlutterGroup(struct=agrd, xfer=rbfi, aero=sdpm, flutter=pypk))

def print_time(selapsed):
    days, rem = divmod(selapsed, 24*60*60)
    hours, rem = divmod(rem, 60*60)
    minutes, seconds = divmod(rem, 60)
    print('Wall-clock time: {:0>2}-{:0>2}:{:0>2}:{:0>2}'.format(int(days),int(hours),int(minutes),int(seconds)))

if __name__ == "__main__":
    import time
    # Set up the problem
    n_runs, mach, _, _ = get_exp()
    prob = om.Problem(model=Agard(n=n_runs), driver=om.ScipyOptimizeDriver(optimizer='SLSQP', tol=1e-6, disp=True, maxiter=20))
    rec_inc = []
    for i in range(n_runs):
        rec_inc.extend([f'flutter_{i}.f', f'flutter_{i}.g', f'flutter_{i}.g_agg', f'flutter_{i}.flutter.f_speed'])
    prob.model.recording_options['includes'] = rec_inc
    prob.model.add_recorder(om.SqliteRecorder('cases.sql'))
    prob.setup()
    om.n2(prob, show_browser=False, outfile='n2.html')

    # Run
    tic = time.perf_counter()
    prob.run_model()
    toc = time.perf_counter()
    print_time(toc-tic)

    # Print the history
    cases = om.CaseReader('cases.sql').get_cases()
    hdr = '{:>10s}'.format('it')
    for irun in range(n_runs):
        hdr += ', {:>10s}, {:>10s}'.format(f'g_agg_{irun}', f'u_f_{irun}')
    print(hdr)
    uf = np.zeros(n_runs)
    for i, case in enumerate(cases):
        val = '{:10d}'.format(i)
        for irun in range(n_runs):
            g_agg = case.get_val(f'flutter_{irun}.g_agg')[0]
            uf[irun] = case.get_val(f'flutter_{irun}.flutter.f_speed')[0]
            val += ', {:10.4f}, {:10.4f}'.format(g_agg, uf[irun])
        print(val)
    # Save results
    np.savetxt('mach_uf.dat', np.column_stack((mach, uf)), delimiter=', ', fmt='%15.2f', header='{:>13s}, {:>15s}'.format('Mach number', 'Flutter speed'))
